@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (c) 2014-2018 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2014-2018, 2020 Claudiu Popa <pcmanticore@gmail.com>
 # Copyright (c) 2014 Google, Inc.
 # Copyright (c) 2014 Denis Laxalde <denis.laxalde@logilab.fr>
 # Copyright (c) 2014 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
@@ -9,9 +9,15 @@
 # Copyright (c) 2016 Derek Gustafson <degustaf@gmail.com>
 # Copyright (c) 2016 Jakub Wilk <jwilk@jwilk.net>
 # Copyright (c) 2016 Ceridwen <ceridwenv@gmail.com>
+# Copyright (c) 2018 Ville Skyttä <ville.skytta@iki.fi>
 # Copyright (c) 2018 Mario Corchero <mcorcherojim@bloomberg.net>
 # Copyright (c) 2018 Mario Corchero <mariocj89@gmail.com>
 # Copyright (c) 2018 Anthony Sottile <asottile@umich.edu>
+# Copyright (c) 2019 Hugo van Kemenade <hugovk@users.noreply.github.com>
+# Copyright (c) 2019 markmcclain <markmcclain@users.noreply.github.com>
+# Copyright (c) 2019 BasPH <BasPH@users.noreply.github.com>
+# Copyright (c) 2020 hippo91 <guillaume.peillex@gmail.com>
+# Copyright (c) 2020 Peter Kolbus <peter.kolbus@gmail.com>
 
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 # For details: https://github.com/PyCQA/astroid/blob/master/COPYING.LESSER
@@ -27,14 +33,16 @@
 :type BUILTIN_MODULES: dict
 :var BUILTIN_MODULES: dictionary with builtin module names has key
 """
-import imp
+import importlib.util
 import os
 import platform
 import sys
 import itertools
-from distutils.sysconfig import get_python_lib # pylint: disable=import-error
+from distutils.sysconfig import get_python_lib  # pylint: disable=import-error
+
 # pylint: disable=import-error, no-name-in-module
 from distutils.errors import DistutilsPlatformError
+
 # distutils is replaced by virtualenv with a module that does
 # weird path manipulations in order to get to the
 # real distutils module.
@@ -42,12 +50,12 @@ from distutils.errors import DistutilsPlatformError
 from .interpreter._import import spec
 from .interpreter._import import util
 
-if sys.platform.startswith('win'):
-    PY_SOURCE_EXTS = ('py', 'pyw')
-    PY_COMPILED_EXTS = ('dll', 'pyd')
+if sys.platform.startswith("win"):
+    PY_SOURCE_EXTS = ("py", "pyw")
+    PY_COMPILED_EXTS = ("dll", "pyd")
 else:
-    PY_SOURCE_EXTS = ('py',)
-    PY_COMPILED_EXTS = ('so',)
+    PY_SOURCE_EXTS = ("py",)
+    PY_COMPILED_EXTS = ("so",)
 
 
 try:
@@ -60,37 +68,50 @@ try:
         get_python_lib(standard_lib=True, prefix=sys.prefix),
         # Take care of installations where exec_prefix != prefix.
         get_python_lib(standard_lib=True, prefix=sys.exec_prefix),
-        get_python_lib(standard_lib=True)}
+        get_python_lib(standard_lib=True),
+    }
 # get_python_lib(standard_lib=1) is not available on pypy, set STD_LIB_DIR to
 # non-valid path, see https://bugs.pypy.org/issue1164
 except DistutilsPlatformError:
     STD_LIB_DIRS = set()
 
-if os.name == 'nt':
-    STD_LIB_DIRS.add(os.path.join(sys.prefix, 'dlls'))
+if os.name == "nt":
+    STD_LIB_DIRS.add(os.path.join(sys.prefix, "dlls"))
     try:
         # real_prefix is defined when running inside virtual environments,
         # created with the **virtualenv** library.
-        STD_LIB_DIRS.add(os.path.join(sys.real_prefix, 'dlls'))
+        STD_LIB_DIRS.add(os.path.join(sys.real_prefix, "dlls"))
     except AttributeError:
         # sys.base_exec_prefix is always defined, but in a virtual environment
         # created with the stdlib **venv** module, it points to the original
         # installation, if the virtual env is activated.
         try:
-            STD_LIB_DIRS.add(os.path.join(sys.base_exec_prefix, 'dlls'))
+            STD_LIB_DIRS.add(os.path.join(sys.base_exec_prefix, "dlls"))
         except AttributeError:
             pass
 
-if platform.python_implementation() == 'PyPy':
-    _root = os.path.join(sys.prefix, 'lib_pypy')
+if platform.python_implementation() == "PyPy":
+    # The get_python_lib(standard_lib=True) function does not give valid
+    # result with pypy in a virtualenv.
+    # In a virtual environment, with CPython implementation the call to this function returns a path toward
+    # the binary (its libraries) which has been used to create the virtual environment.
+    # Not with pypy implementation.
+    # The only way to retrieve such information is to use the sys.base_prefix hint.
+    # It's worth noticing that under CPython implementation the return values of
+    # get_python_lib(standard_lib=True) and get_python_lib(santdard_lib=True, prefix=sys.base_prefix)
+    # are the same.
+    # In the lines above, we could have replace the call to get_python_lib(standard=True)
+    # with the one using prefix=sys.base_prefix but we prefer modifying only what deals with pypy.
+    STD_LIB_DIRS.add(get_python_lib(standard_lib=True, prefix=sys.base_prefix))
+    _root = os.path.join(sys.prefix, "lib_pypy")
     STD_LIB_DIRS.add(_root)
     try:
         # real_prefix is defined when running inside virtualenv.
-        STD_LIB_DIRS.add(os.path.join(sys.real_prefix, 'lib_pypy'))
+        STD_LIB_DIRS.add(os.path.join(sys.base_prefix, "lib_pypy"))
     except AttributeError:
         pass
     del _root
-if os.name == 'posix':
+if os.name == "posix":
     # Need the real prefix is we're under a virtualenv, otherwise
     # the usual one will do.
     try:
@@ -99,11 +120,11 @@ if os.name == 'posix':
         prefix = sys.prefix
 
     def _posix_path(path):
-        base_python = 'python%d.%d' % sys.version_info[:2]
+        base_python = "python%d.%d" % sys.version_info[:2]
         return os.path.join(prefix, path, base_python)
 
-    STD_LIB_DIRS.add(_posix_path('lib'))
-    if sys.maxsize > 2**32:
+    STD_LIB_DIRS.add(_posix_path("lib"))
+    if sys.maxsize > 2 ** 32:
         # This tries to fix a problem with /usr/lib64 builds,
         # where systems are running both 32-bit and 64-bit code
         # on the same machine, which reflects into the places where
@@ -111,10 +132,10 @@ if os.name == 'posix':
         # here http://bugs.python.org/issue1294959.
         # An easy reproducing case would be
         # https://github.com/PyCQA/pylint/issues/712#issuecomment-163178753
-        STD_LIB_DIRS.add(_posix_path('lib64'))
+        STD_LIB_DIRS.add(_posix_path("lib64"))
 
-EXT_LIB_DIR = get_python_lib()
-IS_JYTHON = platform.python_implementation() == 'Jython'
+EXT_LIB_DIRS = {get_python_lib(), get_python_lib(True)}
+IS_JYTHON = platform.python_implementation() == "Jython"
 BUILTIN_MODULES = dict.fromkeys(sys.builtin_module_names, True)
 
 
@@ -123,6 +144,7 @@ class NoSourceFile(Exception):
     source file for a precompiled file
     """
 
+
 def _normalize_path(path):
     return os.path.normcase(os.path.abspath(path))
 
@@ -133,10 +155,6 @@ def _canonicalize_path(path):
 
 def _path_from_filename(filename, is_jython=IS_JYTHON):
     if not is_jython:
-        if sys.version_info > (3, 0):
-            return filename
-        if filename.endswith(".pyc"):
-            return filename[:-1]
         return filename
     head, has_pyclass, _ = filename.partition("$py.class")
     if has_pyclass:
@@ -158,6 +176,7 @@ def _handle_blacklist(blacklist, dirnames, filenames):
 
 _NORM_PATH_CACHE = {}
 
+
 def _cache_normalize_path(path):
     """abspath with caching"""
     # _module_file calls abspath on every path in sys.path every time it's
@@ -166,116 +185,59 @@ def _cache_normalize_path(path):
     try:
         return _NORM_PATH_CACHE[path]
     except KeyError:
-        if not path: # don't cache result for ''
+        if not path:  # don't cache result for ''
             return _normalize_path(path)
         result = _NORM_PATH_CACHE[path] = _normalize_path(path)
         return result
 
-def load_module_from_name(dotted_name, path=None, use_sys=True):
+
+def load_module_from_name(dotted_name):
     """Load a Python module from its name.
 
     :type dotted_name: str
     :param dotted_name: python name of a module or package
 
-    :type path: list or None
-    :param path:
-      optional list of path where the module or package should be
-      searched (use sys.path if nothing or None is given)
-
-    :type use_sys: bool
-    :param use_sys:
-      boolean indicating whether the sys.modules dictionary should be
-      used or not
-
-
     :raise ImportError: if the module or package is not found
 
     :rtype: module
     :return: the loaded module
     """
-    return load_module_from_modpath(dotted_name.split('.'), path, use_sys)
+    try:
+        return sys.modules[dotted_name]
+    except KeyError:
+        pass
+
+    return importlib.import_module(dotted_name)
 
 
-def load_module_from_modpath(parts, path=None, use_sys=1):
+def load_module_from_modpath(parts):
     """Load a python module from its split name.
 
     :type parts: list(str) or tuple(str)
     :param parts:
       python name of a module or package split on '.'
 
-    :type path: list or None
-    :param path:
-      optional list of path where the module or package should be
-      searched (use sys.path if nothing or None is given)
-
-    :type use_sys: bool
-    :param use_sys:
-      boolean indicating whether the sys.modules dictionary should be used or not
-
     :raise ImportError: if the module or package is not found
 
     :rtype: module
     :return: the loaded module
     """
-    if use_sys:
-        try:
-            return sys.modules['.'.join(parts)]
-        except KeyError:
-            pass
-    modpath = []
-    prevmodule = None
-    for part in parts:
-        modpath.append(part)
-        curname = '.'.join(modpath)
-        module = None
-        if len(modpath) != len(parts):
-            # even with use_sys=False, should try to get outer packages from sys.modules
-            module = sys.modules.get(curname)
-        elif use_sys:
-            # because it may have been indirectly loaded through a parent
-            module = sys.modules.get(curname)
-        if module is None:
-            mp_file, mp_filename, mp_desc = imp.find_module(part, path)
-            module = imp.load_module(curname, mp_file, mp_filename, mp_desc)
-            # mp_file still needs to be closed.
-            if mp_file:
-                mp_file.close()
-        if prevmodule:
-            setattr(prevmodule, part, module)
-        _file = getattr(module, '__file__', '')
-        prevmodule = module
-        if not _file and util.is_namespace(curname):
-            continue
-        if not _file and len(modpath) != len(parts):
-            raise ImportError('no module in %s' % '.'.join(parts[len(modpath):]))
-        path = [os.path.dirname(_file)]
-    return module
+    return load_module_from_name(".".join(parts))
 
 
-def load_module_from_file(filepath, path=None, use_sys=True, extrapath=None):
+def load_module_from_file(filepath: str):
     """Load a Python module from it's path.
 
     :type filepath: str
     :param filepath: path to the python module or package
 
-    :type path: list or None
-    :param path:
-      optional list of path where the module or package should be
-      searched (use sys.path if nothing or None is given)
-
-    :type use_sys: bool
-    :param use_sys:
-      boolean indicating whether the sys.modules dictionary should be
-      used or not
-
-
     :raise ImportError: if the module or package is not found
 
     :rtype: module
     :return: the loaded module
     """
-    modpath = modpath_from_file(filepath, extrapath)
-    return load_module_from_modpath(modpath, path, use_sys)
+    modpath = modpath_from_file(filepath)
+    return load_module_from_modpath(modpath)
 
 
 def check_modpath_has_init(path, mod_path):
@@ -285,7 +247,7 @@ def check_modpath_has_init(path, mod_path):
         modpath.append(part)
         path = os.path.join(path, part)
         if not _has_init(path):
-            old_namespace = util.is_namespace('.'.join(modpath))
+            old_namespace = util.is_namespace(".".join(modpath))
             if not old_namespace:
                 return False
     return True
@@ -319,55 +281,42 @@ def _get_relative_base_path(filename, path_to_check):
 
     if importable_path:
         base_path = os.path.splitext(importable_path)[0]
-        relative_base_path = base_path[len(path_to_check):]
+        relative_base_path = base_path[len(path_to_check) :]
         return [pkg for pkg in relative_base_path.split(os.sep) if pkg]
 
     return None
 
 
-def modpath_from_file_with_callback(filename, extrapath=None, is_package_cb=None):
+def modpath_from_file_with_callback(filename, path=None, is_package_cb=None):
     filename = os.path.expanduser(_path_from_filename(filename))
-
-    if extrapath is not None:
-        for path_ in itertools.chain(map(_canonicalize_path, extrapath),
-                                     extrapath):
-            path = os.path.abspath(path_)
-            if not path:
-                continue
-            submodpath = _get_relative_base_path(filename, path)
-            if not submodpath:
-                continue
-            if is_package_cb(path, submodpath[:-1]):
-                return extrapath[path_].split('.') + submodpath
-
-    for path in itertools.chain(map(_canonicalize_path, sys.path), sys.path):
-        path = _cache_normalize_path(path)
-        if not path:
+    for pathname in itertools.chain(
+        path or [], map(_canonicalize_path, sys.path), sys.path
+    ):
+        pathname = _cache_normalize_path(pathname)
+        if not pathname:
             continue
-        modpath = _get_relative_base_path(filename, path)
+        modpath = _get_relative_base_path(filename, pathname)
         if not modpath:
             continue
-        if is_package_cb(path, modpath[:-1]):
+        if is_package_cb(pathname, modpath[:-1]):
             return modpath
 
-    raise ImportError('Unable to find module for %s in %s' % (
-        filename, ', \n'.join(sys.path)))
+    raise ImportError(
+        "Unable to find module for %s in %s" % (filename, ", \n".join(sys.path))
+    )
 
 
+def modpath_from_file(filename, path=None):
+    """Get the corresponding split module's name from a filename
 
-def modpath_from_file(filename, extrapath=None):
-    """given a file path return the corresponding split module's name
-    (i.e name of a module or package split on '.')
+    This function will return the name of a module or package split on `.`.
 
     :type filename: str
     :param filename: file's path for which we want the module's name
 
-    :type extrapath: dict
-    :param extrapath:
-      optional extra search path, with path as key and package name for the path
-      as value. This is usually useful to handle package split in multiple
-      directories using __path__ trick.
-
+    :type Optional[List[str]] path:
+      Optional list of path where the module or package should be
+      searched (use sys.path if nothing or None is given)
 
     :raise ImportError:
       if the corresponding module's name has not been found
@@ -375,12 +324,13 @@ def modpath_from_file(filename, extrapath=None):
     :rtype: list(str)
     :return: the corresponding split module's name
     """
-    return modpath_from_file_with_callback(filename, extrapath, check_modpath_has_init)
+    return modpath_from_file_with_callback(filename, path, check_modpath_has_init)
 
 
 def file_from_modpath(modpath, path=None, context_file=None):
     return file_info_from_modpath(modpath, path, context_file).location
 
+
 def file_info_from_modpath(modpath, path=None, context_file=None):
     """given a mod path (i.e. split module / package name), return the
     corresponding file, giving priority to source file over precompiled
@@ -415,15 +365,19 @@ def file_info_from_modpath(modpath, path=None, context_file=None):
         context = os.path.dirname(context_file)
     else:
         context = context_file
-    if modpath[0] == 'xml':
+    if modpath[0] == "xml":
         # handle _xmlplus
         try:
-            return _spec_from_modpath(['_xmlplus'] + modpath[1:], path, context)
+            return _spec_from_modpath(["_xmlplus"] + modpath[1:], path, context)
         except ImportError:
             return _spec_from_modpath(modpath, path, context)
-    elif modpath == ['os', 'path']:
+    elif modpath == ["os", "path"]:
         # FIXME: currently ignoring search_path...
-        return spec.ModuleSpec(name='os.path', location=os.path.__file__, module_type=imp.PY_SOURCE)
+        return spec.ModuleSpec(
+            name="os.path",
+            location=os.path.__file__,
+            module_type=spec.ModuleType.PY_SOURCE,
+        )
     return _spec_from_modpath(modpath, path, context)
 
 
@@ -454,9 +408,9 @@ def get_module_part(dotted_name, context_file=None):
     (see #10066)
     """
     # os.path trick
-    if dotted_name.startswith('os.path'):
-        return 'os.path'
-    parts = dotted_name.split('.')
+    if dotted_name.startswith("os.path"):
+        return "os.path"
+    parts = dotted_name.split(".")
     if context_file is not None:
         # first check for builtin module which won't be considered latter
         # in that case (path != None)
@@ -467,22 +421,24 @@ def get_module_part(dotted_name, context_file=None):
         # don't use += or insert, we want a new list to be created !
     path = None
     starti = 0
-    if parts[0] == '':
-        assert context_file is not None, \
-                'explicit relative import, but no context_file?'
-        path = [] # prevent resolving the import non-relatively
+    if parts[0] == "":
+        assert (
+            context_file is not None
+        ), "explicit relative import, but no context_file?"
+        path = []  # prevent resolving the import non-relatively
         starti = 1
-    while parts[starti] == '': # for all further dots: change context
+    while parts[starti] == "":  # for all further dots: change context
         starti += 1
         context_file = os.path.dirname(context_file)
     for i in range(starti, len(parts)):
         try:
-            file_from_modpath(parts[starti:i+1], path=path,
-                              context_file=context_file)
+            file_from_modpath(
+                parts[starti : i + 1], path=path, context_file=context_file
+            )
         except ImportError:
             if i < max(1, len(parts) - 2):
                 raise
-            return '.'.join(parts[:i])
+            return ".".join(parts[:i])
     return dotted_name
 
 
@@ -513,7 +469,7 @@ def get_module_files(src_directory, blacklist, list_all=False):
             continue
         _handle_blacklist(blacklist, dirnames, filenames)
         # check for __init__.py
-        if not list_all and '__init__.py' not in filenames:
+        if not list_all and "__init__.py" not in filenames:
             dirnames[:] = ()
             continue
         for filename in filenames:
@@ -540,7 +496,7 @@ def get_source_file(filename, include_no_ext=False):
     filename = os.path.abspath(_path_from_filename(filename))
     base, orig_ext = os.path.splitext(filename)
     for ext in PY_SOURCE_EXTS:
-        source_path = '%s.%s' % (base, ext)
+        source_path = "%s.%s" % (base, ext)
         if os.path.exists(source_path):
             return source_path
     if include_no_ext and not orig_ext and os.path.exists(base):
@@ -573,7 +529,7 @@ def is_standard_module(modname, std_path=None):
       - is located on the path listed in one of the directory in `std_path`
       - is a built-in module
     """
-    modname = modname.split('.')[0]
+    modname = modname.split(".")[0]
     try:
         filename = file_from_modpath([modname])
     except ImportError:
@@ -586,8 +542,9 @@ def is_standard_module(modname, std_path=None):
         # we assume there are no namespaces in stdlib
         return not util.is_namespace(modname)
     filename = _normalize_path(filename)
-    if filename.startswith(_cache_normalize_path(EXT_LIB_DIR)):
-        return False
+    for path in EXT_LIB_DIRS:
+        if filename.startswith(_cache_normalize_path(path)):
+            return False
     if std_path is None:
         std_path = STD_LIB_DIRS
     for path in std_path:
@@ -596,7 +553,6 @@ def is_standard_module(modname, std_path=None):
     return False
 
 
-
 def is_relative(modname, from_file):
     """return true if the given module name is relative to the given
     file name
@@ -616,19 +572,26 @@ def is_relative(modname, from_file):
         from_file = os.path.dirname(from_file)
     if from_file in sys.path:
         return False
-    try:
-        stream, _, _ = imp.find_module(modname.split('.')[0], [from_file])
+    name = os.path.basename(from_file)
+    file_path = os.path.dirname(from_file)
+    parent_spec = importlib.util.find_spec(name, from_file)
+    while parent_spec is None and len(file_path) > 0:
+        name = os.path.basename(file_path) + "." + name
+        file_path = os.path.dirname(file_path)
+        parent_spec = importlib.util.find_spec(name, from_file)
 
-        # Close the stream to avoid ResourceWarnings.
-        if stream:
-            stream.close()
-        return True
-    except ImportError:
+    if parent_spec is None:
         return False
 
+    submodule_spec = importlib.util.find_spec(
+        name + "." + modname.split(".")[0], parent_spec.submodule_search_locations
+    )
+    return submodule_spec is not None
+
 
 # internal only functions #####################################################
 
+
 def _spec_from_modpath(modpath, path=None, context=None):
     """given a mod path (i.e. split module / package name), return the
     corresponding spec
@@ -650,7 +613,9 @@ def _spec_from_modpath(modpath, path=None, context=None):
     if found_spec.type == spec.ModuleType.PY_COMPILED:
         try:
             location = get_source_file(found_spec.location)
-            return found_spec._replace(location=location, type=spec.ModuleType.PY_SOURCE)
+            return found_spec._replace(
+                location=location, type=spec.ModuleType.PY_SOURCE
+            )
         except NoSourceFile:
             return found_spec._replace(location=location)
     elif found_spec.type == spec.ModuleType.C_BUILTIN:
@@ -667,25 +632,23 @@ def _is_python_file(filename):
 
     .pyc and .pyo are ignored
     """
-    for ext in ('.py', '.so', '.pyd', '.pyw'):
-        if filename.endswith(ext):
-            return True
-    return False
+    return filename.endswith((".py", ".so", ".pyd", ".pyw"))
 
 
 def _has_init(directory):
     """if the given directory has a valid __init__ file, return its path,
     else return None
     """
-    mod_or_pack = os.path.join(directory, '__init__')
-    for ext in PY_SOURCE_EXTS + ('pyc', 'pyo'):
-        if os.path.exists(mod_or_pack + '.' + ext):
-            return mod_or_pack + '.' + ext
+    mod_or_pack = os.path.join(directory, "__init__")
+    for ext in PY_SOURCE_EXTS + ("pyc", "pyo"):
+        if os.path.exists(mod_or_pack + "." + ext):
+            return mod_or_pack + "." + ext
     return None
 
 
 def is_namespace(specobj):
     return specobj.type == spec.ModuleType.PY_NAMESPACE
 
+
 def is_directory(specobj):
     return specobj.type == spec.ModuleType.PKG_DIRECTORY
